/*************************************************************************
* (C) 2008 AITNET ltd - Sofia/Bulgaria - <misho@aitbg.com>
*  by Michael Pounov <misho@openbsd-bg.org>
*
* $Author: misho $
* $Id: aitcfg.c,v 1.12 2014/03/03 09:41:09 misho Exp $
*
**************************************************************************
The ELWIX and AITNET software is distributed under the following
terms:

All of the documentation and software included in the ELWIX and AITNET
Releases is copyrighted by ELWIX - Sofia/Bulgaria <info@elwix.org>

Copyright 2004 - 2014
	by Michael Pounov <misho@elwix.org>.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
This product includes software developed by Michael Pounov <misho@elwix.org>
ELWIX - Embedded LightWeight unIX and its contributors.
4. Neither the name of AITNET nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AITNET AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/
#include "global.h"


#pragma GCC visibility push(hidden)

int cfg_Errno;
char cfg_Error[STRSIZ];

int
cfg_Write(FILE *f, char *fmt, ...)
{
	int ret = 0;
	va_list lst;

	va_start(lst, fmt);
	ret = vfprintf(f, fmt, lst);
	va_end(lst);

	return ret;
}

int
cfg_tree_cmp(struct tagCfg *a, struct tagCfg *b)
{
	int ret;

	assert(a && b);

	ret = ((AIT_KEY(&a->cfg_sec) << 16) | AIT_KEY(&a->cfg_attr)) - 
		((AIT_KEY(&b->cfg_sec) << 16) | AIT_KEY(&b->cfg_attr));

	if (ret < 0)
		return -1;
	else if (ret > 0)
		return 1;

	return ret;
}

RB_GENERATE(tagRC, tagCfg, cfg_node, cfg_tree_cmp);

#pragma GCC visibility pop


// cfg_GetErrno() Get error code of last operation
int
cfg_GetErrno()
{
	return cfg_Errno;
}

// cfg_GetError() Get error text of last operation
const char *
cfg_GetError()
{
	return cfg_Error;
}

// cfg_SetErr() Set error to variables for internal use!!!
void
cfg_SetErr(int eno, char *estr, ...)
{
	va_list lst;

	cfg_Errno = eno;
	memset(cfg_Error, 0, sizeof cfg_Error);
	va_start(lst, estr);
	vsnprintf(cfg_Error, sizeof cfg_Error, estr, lst);
	va_end(lst);
}


/*
 * cfgInitConfig() - Init config root
 *
 * return: NULL error or !=NULL allocated config root
 */
cfg_root_t *
cfgInitConfig()
{
	cfg_root_t *cfg = NULL;

	cfg = e_malloc(sizeof(cfg_root_t));
	if (!cfg) {
		cfg_SetErr(elwix_GetErrno(), "%s", elwix_GetError());
		return NULL;
	} else
		memset(cfg, 0, sizeof(cfg_root_t));

	pthread_mutex_init(&cfg->rc_mtx, NULL);

	TAILQ_INIT(cfg);
	RB_INIT(cfg);
	return cfg;
}

/*
 * cfgEndConfig() - Free resources & config root
 *
 * @pcfg = Config root
 * return: none
 */
void
cfgEndConfig(cfg_root_t **pcfg)
{
	if (pcfg && *pcfg) {
		cfgClearConfig(*pcfg);
		pthread_mutex_destroy(&(*pcfg)->rc_mtx);
		e_free(*pcfg);
		*pcfg = NULL;
	}
}

/*
 * cfgLoadConfig() - Load config from file
 *
 * @cfgName = Config filename
 * @cfg = Config root
 * return: -1 error or 0 ok
 */
int
cfgLoadConfig(const char *cfgName, cfg_root_t * __restrict cfg)
{
	FILE *f;
	int ret;

	if (!cfgName || !cfg) {
		cfg_SetErr(EINVAL, "Invalid parameter(s)");
		return -1;
	} else {
		pthread_mutex_init(&cfg->rc_mtx, NULL);

		TAILQ_INIT(cfg);
		RB_INIT(cfg);
	}

	f = fopen(cfgName, "r");
	if (!f) {
		LOGERR;
		return -1;
	}

	ret = cfgReadConfig(f, cfg);

	fclose(f);
	return ret;
}

/*
 * cfgClearConfig() - Clear config and free resources
 *
 * @cfg = Config root
 * return: none
 */
void
cfgClearConfig(cfg_root_t * __restrict cfg)
{
	struct tagCfg *av;

	if (!cfg)
		return;

	CFG_RC_LOCK(cfg);
	while ((av = TAILQ_FIRST(cfg))) {
		TAILQ_REMOVE(cfg, av, cfg_next);

		AIT_FREE_VAL(&av->cfg_val);
		AIT_FREE_VAL(&av->cfg_attr);
		AIT_FREE_VAL(&av->cfg_sec);
		e_free(av);
	}
	cfg->rbh_root = NULL;
	CFG_RC_UNLOCK(cfg);
}

/*
 * cfgUnloadConfig() - Unload config from memory and destroy resources
 *
 * @cfg = Config root
 * return: none
 */
void
cfgUnloadConfig(cfg_root_t * __restrict cfg)
{
	if (!cfg)
		return;

	cfgClearConfig(cfg);
	pthread_mutex_destroy(&cfg->rc_mtx);
}

/*
 * cfgCreateConfig() - Create config file from memory
 *
 * @csConfigName = New config filename
 * @cfg = Config root
 * @whitespace = Additional whitespace characters to file
 * return: -1 error or 0 ok
 */
int
cfgCreateConfig(const char *csConfigName, cfg_root_t * __restrict cfg, int whitespace)
{
	FILE *f;
	int ret;

	if (!csConfigName || !cfg)
		return -1;

	f = fopen(csConfigName, "w");
	if (!f) {
		LOGERR;
		return -1;
	}
	
	ret = cfgWriteConfig(f, cfg, whitespace);

	fclose(f);
	return ret;
}

/*
 * cfgInitPasswd() - Init password root
 *
 * return: NULL error or !=NULL allocated password root
 */
pwd_root_t *
cfgInitPasswd()
{
	pwd_root_t *pwd = NULL;

	pwd = e_malloc(sizeof(pwd_root_t));
	if (!pwd) {
		cfg_SetErr(elwix_GetErrno(), "%s", elwix_GetError());
		return NULL;
	} else
		memset(pwd, 0, sizeof(pwd_root_t));

	pthread_mutex_init(&pwd->pwd_mtx, NULL);

	SLIST_INIT(pwd);
	RB_INIT(pwd);
	return 0;
}

/*
 * cfgEndPasswd() - Free resources & password root
 *
 * @ppwd = Password root
 * return: none
 */
void
cfgEndPasswd(pwd_root_t **ppwd)
{
	if (ppwd && *ppwd) {
		cfgClearPasswd(*ppwd);
		pthread_mutex_destroy(&(*ppwd)->pwd_mtx);
		e_free(*ppwd);
		*ppwd = NULL;
	}
}

/*
 * cfgLoadPasswd() - Load passwords from file
 *
 * @pwdName = Passwords filename
 * @pwd = Password root
 * return: -1 error or 0 ok
 */
int
cfgLoadPasswd(const char *pwdName, pwd_root_t * __restrict pwd)
{
	FILE *f;
	int ret;

	if (!pwdName || !pwd) {
		cfg_SetErr(EINVAL, "Invalid parameter(s)");
		return -1;
	} else {
		pthread_mutex_init(&pwd->pwd_mtx, NULL);

		SLIST_INIT(pwd);
		RB_INIT(pwd);
	}

	f = fopen(pwdName, "r");
	if (!f) {
		LOGERR;
		return -1;
	}

	ret = cfgReadPasswd(f, pwd);

	fclose(f);
	return ret;
}

/*
 * cfgClearPasswd() - Clear passwords and free resources
 *
 * @cfg = Password root
 * return: none
 */
void
cfgClearPasswd(pwd_root_t * __restrict pwd)
{
	struct tagUser *p;

	if (!pwd)
		return;

	PWD_LOCK(pwd);
	while ((p = SLIST_FIRST(pwd))) {
		SLIST_REMOVE_HEAD(pwd, usr_next);

		AIT_FREE_VAL(&p->usr_name);
		AIT_FREE_VAL(&p->usr_pass);
		AIT_FREE_VAL(&p->usr_uid);
		AIT_FREE_VAL(&p->usr_gid);
		AIT_FREE_VAL(&p->usr_class);
		AIT_FREE_VAL(&p->usr_change);
		AIT_FREE_VAL(&p->usr_expire);
		AIT_FREE_VAL(&p->usr_realm);
		AIT_FREE_VAL(&p->usr_home);
		AIT_FREE_VAL(&p->usr_shell);
		e_free(p);
	}
	pwd->rbh_root = NULL;
	PWD_UNLOCK(pwd);
}

/*
 * cfgUnloadPasswd() - Unload passwords from memory and destroy resources
 *
 * @pwd = Password root
 * return: none
 */
void
cfgUnloadPasswd(pwd_root_t * __restrict pwd)
{
	if (!pwd)
		return;

	cfgClearPasswd(pwd);
	pthread_mutex_destroy(&pwd->pwd_mtx);
}

/*
 * cfgCreatePasswd() - Create password file from memory
 *
 * @pwdName = New password filename
 * @pwd = Password root
 * return: -1 error or 0 ok
 */
int
cfgCreatePasswd(const char *pwdName, pwd_root_t * __restrict pwd)
{
	FILE *f;
	int ret;

	if (!pwdName || !pwd)
		return -1;

	f = fopen(pwdName, "w");
	if (!f) {
		LOGERR;
		return -1;
	}
	
	ret = cfgWritePasswd(f, pwd);

	fclose(f);
	return ret;
}
